#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.constants as common_constants

import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.portfolio as portfolio_class


class MarginPortfolio(portfolio_class.Portfolio):
    # parse the exchange balance
    def _parse_currency_balance(self, currency_balance):
        return self._create_currency_portfolio(
            available=currency_balance[constants.CONFIG_PORTFOLIO_FREE]
            if constants.CONFIG_PORTFOLIO_FREE in currency_balance else currency_balance[
                common_constants.PORTFOLIO_AVAILABLE],
            margin=currency_balance[constants.CONFIG_PORTFOLIO_MARGIN]
            if constants.CONFIG_PORTFOLIO_MARGIN in currency_balance else (
                currency_balance[common_constants.MARGIN_PORTFOLIO]
                if common_constants.MARGIN_PORTFOLIO in currency_balance else 0),
            total=currency_balance[constants.CONFIG_PORTFOLIO_TOTAL]
            if constants.CONFIG_PORTFOLIO_TOTAL in currency_balance else currency_balance[
                common_constants.PORTFOLIO_TOTAL])

    def _create_currency_portfolio(self, available, total, margin=0):
        return {common_constants.PORTFOLIO_AVAILABLE: available, common_constants.MARGIN_PORTFOLIO: margin,
                common_constants.PORTFOLIO_TOTAL: total}

    def _reset_currency_portfolio(self, currency):
        self._set_currency_portfolio(currency=currency, available=0, total=0, margin=0)

    def _set_currency_portfolio(self, currency, available, total, margin=0):
        self.portfolio[currency] = self._create_currency_portfolio(available=available, total=total, margin=margin)

    def _update_currency_portfolio(self, currency, available=0, total=0, margin=0):
        self.portfolio[currency][common_constants.PORTFOLIO_AVAILABLE] += available
        self.portfolio[currency][common_constants.MARGIN_PORTFOLIO] += margin
        self.portfolio[currency][common_constants.PORTFOLIO_TOTAL] += total



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_commons.constants as commons_constants

import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.constants as constants
import octobot_trading.personal_data as personal_data
import octobot_trading.util as util


class PortfolioManager(util.Initializable):
    """
    Manage the portfolio and portfolio profitability instances
    """

    def __init__(self, config, trader, exchange_manager):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.config, self.trader, self.exchange_manager = config, trader, exchange_manager

        self.portfolio = None
        self.portfolio_profitability = None
        self.portfolio_value_holder = None
        self.reference_market = None

    async def initialize_impl(self):
        """
        Reset the portfolio instance
        """
        await self._reset_portfolio()

    def handle_balance_update(self, balance, is_diff_update=False):
        """
        Handle a balance update request
        :param balance: the new balance
        :param is_diff_update: True when the update is a partial portfolio
        :return: True if the portfolio was updated
        """
        if self.trader.is_enabled and balance is not None:
            return self.portfolio.update_portfolio_from_balance(balance, force_replace=not is_diff_update)
        return False

    async def handle_balance_update_from_order(self, order) -> bool:
        """
        Handle a balance update from an order request
        :param order: the order
        :return: True if the portfolio was updated
        """
        if self.trader.is_enabled:
            if self.trader.simulate:
                return self._refresh_simulated_trader_portfolio_from_order(order)
            # on real trading: reload portfolio to ensure portfolio sync
            return await self._refresh_real_trader_portfolio()
        return False

    async def handle_balance_updated(self):
        """
        Handle balance update notification
        :return: True if profitability changed
        """
        return await self.portfolio_profitability.update_profitability()

    async def handle_profitability_recalculation(self, force_recompute_origin_portfolio):
        """
        Called before PortfolioProfitability's portfolio profitability recalculation
        to ensure portfolio values are available
        :param force_recompute_origin_portfolio: when True, force origin portfolio computation
        """
        await self.portfolio_value_holder.handle_profitability_recalculation(force_recompute_origin_portfolio)

    async def handle_mark_price_update(self, symbol, mark_price):
        """
        Handle a mark price update notification
        :param symbol: the update symbol
        :param mark_price: the updated mark price
        :return: True if profitability changed
        """
        return await self.portfolio_profitability. \
            update_profitability(force_recompute_origin_portfolio=self.portfolio_value_holder.
                                 update_origin_crypto_currencies_values(symbol, mark_price))

    async def _refresh_real_trader_portfolio(self) -> bool:
        """
        Call BALANCE_CHANNEL producer to refresh real trader portfolio
        :return: True if the portfolio was updated
        """
        return await exchange_channel.get_chan(constants.BALANCE_CHANNEL,
                                               self.exchange_manager.id).get_internal_producer().\
            refresh_real_trader_portfolio()

    async def _reset_portfolio(self):
        """
        Reset the portfolio and portfolio profitability instances
        """
        self.portfolio = personal_data.create_portfolio_from_exchange_manager(self.exchange_manager)
        await self.portfolio.initialize()
        self._load_portfolio()

        self.reference_market = util.get_reference_market(self.config)
        self.portfolio_value_holder = personal_data.PortfolioValueHolder(self)
        self.portfolio_profitability = personal_data.PortfolioProfitability(self)

    def _refresh_simulated_trader_portfolio_from_order(self, order):
        """
        Handle a balance update from an order request when simulating
        :param order: the order that should update portfolio
        :return: True if the portfolio was updated
        """
        if order.is_filled():
            self.portfolio.update_portfolio_from_filled_order(order)
        else:
            self.portfolio.update_portfolio_available(order, is_new_order=False)
        return True

    def _load_portfolio(self):
        """
        Load simulated portfolio from config if required
        """
        if self.trader.is_enabled:
            if self.trader.simulate:
                self._set_starting_simulated_portfolio()
            self.logger.info(f"{constants.CURRENT_PORTFOLIO_STRING} {self.portfolio.portfolio}")

    def _set_starting_simulated_portfolio(self):
        """
        Load new portfolio from config settings
        """
        portfolio_amount_dict = self.config[commons_constants.CONFIG_SIMULATOR][
            commons_constants.CONFIG_STARTING_PORTFOLIO]

        try:
            self.handle_balance_update(self.portfolio.get_portfolio_from_amount_dict(portfolio_amount_dict))
        except Exception as balance_update_exception:
            self.logger.exception(balance_update_exception, True, f"Error when loading trading history, "
                                                                  f"will reset history. ({balance_update_exception})")
            self.handle_balance_update(self.portfolio.get_portfolio_from_amount_dict(
                self.config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]))

    def clear(self):
        """
        Clear portfolio manager objects
        """
        self.portfolio_profitability = None
        self.portfolio_value_holder = None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums
import octobot_trading.personal_data as personal_data


def test_get_min_max_amounts():
    # normal values
    symbol_market = {
        enums.ExchangeConstantsMarketStatusColumns.LIMITS.value: {
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MIN.value: 0.5,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MAX.value: 100,
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MIN.value: None,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MAX.value: None
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MIN.value: 0.5,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MAX.value: 50
            },
        }
    }
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = personal_data.get_min_max_amounts(
        symbol_market)
    assert min_quantity == 0.5
    assert max_quantity == 100
    assert min_cost is None
    assert max_cost is None
    assert min_price == 0.5
    assert max_price == 50

    # missing all values
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = personal_data.get_min_max_amounts({})
    assert min_quantity is None
    assert max_quantity is None
    assert min_cost is None
    assert max_cost is None
    assert min_price is None
    assert max_price is None

    # missing all values: asign default
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = personal_data.get_min_max_amounts({}, "xyz")
    assert min_quantity == "xyz"
    assert max_quantity == "xyz"
    assert min_cost == "xyz"
    assert max_cost == "xyz"
    assert min_price == "xyz"
    assert max_price == "xyz"

    # missing values: assign default

    symbol_market = {
        enums.ExchangeConstantsMarketStatusColumns.LIMITS.value: {
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MIN.value: 0.5,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MAX.value: 100,
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MIN.value: None,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MAX.value: None
            }
        }
    }
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = personal_data.get_min_max_amounts(symbol_market, "xyz")
    assert min_quantity == 0.5
    assert max_quantity == 100
    assert min_cost == "xyz"  # None is not a valid value => assign default
    assert max_cost == "xyz"  # None is not a valid value => assign default
    assert min_price == "xyz"
    assert max_price == "xyz"


def test_get_fees_for_currency():
    fee1 = {
        enums.FeePropertyColumns.CURRENCY.value: "BTC",
        enums.FeePropertyColumns.COST.value: 1
    }
    assert personal_data.get_fees_for_currency(fee1, "BTC") == 1
    assert personal_data.get_fees_for_currency(fee1, "BTC1") == 0

    fee2 = {
        enums.FeePropertyColumns.CURRENCY.value: "BTC",
        enums.FeePropertyColumns.COST.value: 0
    }
    assert personal_data.get_fees_for_currency(fee2, "BTC") == 0
    assert personal_data.get_fees_for_currency(fee2, "BTC1") == 0

    assert personal_data.get_fees_for_currency({}, "BTC") == 0
    assert personal_data.get_fees_for_currency(None, "BTC") == 0



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.constants as commons_constants
import octobot_trading.constants as trading_constants
import octobot_commons.symbol_util as symbol_util


def is_trader_enabled(config) -> bool:
    return _is_trader_enabled(config, commons_constants.CONFIG_TRADER)


def is_trader_simulator_enabled(config) -> bool:
    return _is_trader_enabled(config, commons_constants.CONFIG_SIMULATOR)


def _is_trader_enabled(config, trader_key) -> bool:
    try:
        return config[trader_key][commons_constants.CONFIG_ENABLED_OPTION]
    except KeyError:
        if trader_key not in config:
            config[trader_key] = {}
        config[trader_key][commons_constants.CONFIG_ENABLED_OPTION] = False
        return False


def is_trade_history_loading_enabled(config, default=True) -> bool:
    try:
        return config[commons_constants.CONFIG_TRADER].get(commons_constants.CONFIG_LOAD_TRADE_HISTORY, default)
    except KeyError:
        return default


def is_currency_enabled(config, currency, default_value) -> bool:
    return config[commons_constants.CONFIG_CRYPTO_CURRENCIES][currency].get(commons_constants.CONFIG_ENABLED_OPTION,
                                                                            default_value)


def get_symbols(config, enabled_only) -> list:
    if commons_constants.CONFIG_CRYPTO_CURRENCIES in config \
            and isinstance(config[commons_constants.CONFIG_CRYPTO_CURRENCIES], dict):
        return [
            symbol
            for currency, crypto_currency_data in config[commons_constants.CONFIG_CRYPTO_CURRENCIES].items()
            if not enabled_only or is_currency_enabled(config, currency, True)
            for symbol in crypto_currency_data[commons_constants.CONFIG_CRYPTO_PAIRS]
        ]
    return []


def get_all_currencies(config, enabled_only=False) -> set:
    currencies = set()
    for symbol in get_symbols(config, enabled_only):
        quote, base = symbol_util.split_symbol(symbol)
        currencies.add(quote)
        currencies.add(base)
    return currencies


def get_pairs(config, currency, enabled_only=False) -> list:
    return [
        symbol
        for symbol in get_symbols(config, enabled_only)
        if currency in symbol_util.split_symbol(symbol)
    ]


def get_market_pair(config, currency, enabled_only=False) -> (str, bool):
    if commons_constants.CONFIG_TRADING in config:
        reference_market = get_reference_market(config)
        for symbol in get_symbols(config, enabled_only):
            symbol_currency, symbol_market = symbol_util.split_symbol(symbol)
            if currency == symbol_currency and reference_market == symbol_market:
                return symbol, False
            elif reference_market == symbol_currency and currency == symbol_market:
                return symbol, True
    return "", False


def get_reference_market(config) -> str:
    # The reference market is the currency unit of the calculated quantity value
    return config[commons_constants.CONFIG_TRADING].get(commons_constants.CONFIG_TRADER_REFERENCE_MARKET,
                                                        trading_constants.DEFAULT_REFERENCE_MARKET)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import collections

import octobot_commons.logging as logging

import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums
import octobot_trading.util as util


class PositionsManager(util.Initializable):
    MAX_POSITIONS_COUNT = 2000

    def __init__(self, trader):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.trader = trader
        self.positions_initialized = False  # TODO
        self.positions = collections.OrderedDict()

    async def initialize_impl(self):
        self._reset_positions()

    def get_open_positions(self, symbol=None, since=-1, limit=-1):
        return self._select_positions(True, symbol, since, limit)

    def get_closed_positions(self, symbol=None, since=-1, limit=-1):
        return self._select_positions(False, symbol, since, limit)

    def upsert_position(self, position_id, raw_position) -> bool:
        if position_id not in self.positions:
            self.positions[position_id] = self._create_position_from_raw(raw_position)
            self._check_positions_size()
            return True

        return self._update_position_from_raw(self.positions[position_id], raw_position)

    def upsert_position_instance(self, position) -> bool:
        if position.position_id not in self.positions:
            self.positions[position.position_id] = position
            self._check_positions_size()
            return True
        # TODO
        return False

    # private
    def _check_positions_size(self):
        if len(self.positions) > self.MAX_POSITIONS_COUNT:
            self._remove_oldest_positions(int(self.MAX_POSITIONS_COUNT / 2))

    def _create_position_from_raw(self, raw_position):
        position = personal_data.Position(self.trader)
        position.update_position_from_raw(raw_position)
        return position

    def _update_position_from_raw(self, position, raw_position):
        return position.update_position_from_raw(raw_position)

    def _select_positions(self, status=enums.PositionStatus.OPEN, symbol=None, since=-1, limit=-1):
        positions = [
            position
            for position in self.positions.values()
            if (
                    position.status == status and
                    (symbol is None or (symbol and position.symbol == symbol)) and
                    (since == -1 or (since and position.timestamp < since))
            )
        ]
        return positions if limit == -1 else positions[0:limit]

    def _reset_positions(self):
        self.positions_initialized = False
        self.positions = collections.OrderedDict()

    def _remove_oldest_positions(self, nb_to_remove):
        for _ in range(nb_to_remove):
            self.positions.popitem(last=False)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.symbol_util as symbol_util
import octobot_commons.logging as logging
import octobot_trading.enums as enums
import octobot_trading.exchanges.util.exchange_market_status_fixer as exchange_market_status_fixer
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc


def is_valid(element, key):
    """
    Checks is the element is valid with the market status fixer
    :param element:
    :param key:
    :return:
    """
    return key in element and exchange_market_status_fixer.is_ms_valid(element[key])


def get_min_max_amounts(symbol_market, default_value=None):
    """
    Returns the min and max quantity, cost and price according to the specified market
    :param symbol_market:
    :param default_value:
    :return:
    """
    min_quantity = max_quantity = min_cost = max_cost = min_price = max_price = default_value
    if Ecmsc.LIMITS.value in symbol_market:
        symbol_market_limits = symbol_market[Ecmsc.LIMITS.value]

        if Ecmsc.LIMITS_AMOUNT.value in symbol_market_limits:
            limit_amount = symbol_market_limits[Ecmsc.LIMITS_AMOUNT.value]
            if is_valid(limit_amount, Ecmsc.LIMITS_AMOUNT_MIN.value) \
                    or is_valid(limit_amount, Ecmsc.LIMITS_AMOUNT_MAX.value):
                min_quantity = limit_amount.get(Ecmsc.LIMITS_AMOUNT_MIN.value, default_value)
                max_quantity = limit_amount.get(Ecmsc.LIMITS_AMOUNT_MAX.value, default_value)

        # case 2: use cost and price
        if Ecmsc.LIMITS_COST.value in symbol_market_limits:
            limit_cost = symbol_market_limits[Ecmsc.LIMITS_COST.value]
            if is_valid(limit_cost, Ecmsc.LIMITS_COST_MIN.value) \
                    or is_valid(limit_cost, Ecmsc.LIMITS_COST_MAX.value):
                min_cost = limit_cost.get(Ecmsc.LIMITS_COST_MIN.value, default_value)
                max_cost = limit_cost.get(Ecmsc.LIMITS_COST_MAX.value, default_value)

        # case 2: use quantity and price
        if Ecmsc.LIMITS_PRICE.value in symbol_market_limits:
            limit_price = symbol_market_limits[Ecmsc.LIMITS_PRICE.value]
            if is_valid(limit_price, Ecmsc.LIMITS_PRICE_MIN.value) \
                    or is_valid(limit_price, Ecmsc.LIMITS_PRICE_MAX.value):
                min_price = limit_price.get(Ecmsc.LIMITS_PRICE_MIN.value, default_value)
                max_price = limit_price.get(Ecmsc.LIMITS_PRICE_MAX.value, default_value)

    return min_quantity, max_quantity, min_cost, max_cost, min_price, max_price


def check_cost(total_order_price, min_cost):
    """
    Checks and adapts the quantity and price of the order to ensure it's exchange compliant:
    - are the quantity and price of the order compliant with the exchange's number of digits requirement
        => otherwise quantity will be truncated accordingly
    - is the quantity valid
    - are the order total price and quantity superior or equal to the exchange's minimum order requirement
        => otherwise order is impossible => returns empty list
    - if total cost data are unavailable:
    - is the price of the currency compliant with the exchange's price interval for this currency
        => otherwise order is impossible => returns empty list
    - are the order total price and quantity inferior or equal to the exchange's maximum order requirement
        => otherwise order is impossible as is => split order into smaller ones and returns the list
    => returns the quantity and price list of possible order(s)
    - if exchange symbol data are not enough
        => try fixing exchange data using ExchangeMarketStatusFixer are start again (once only)
    """
    if total_order_price < min_cost:
        if min_cost is None:
            logging.get_logger().error("Invalid min_cost from exchange")
        return False
    return True


async def get_pre_order_data(exchange_manager, symbol: str, timeout: int = None):
    try:
        mark_price = await exchange_manager.exchange_symbols_data.get_exchange_symbol_data(symbol) \
            .prices_manager.get_mark_price(timeout=timeout)
    except asyncio.TimeoutError:
        raise asyncio.TimeoutError("Mark price is not available")

    currency, market = symbol_util.split_symbol(symbol)

    current_symbol_holding = exchange_manager.exchange_personal_data.portfolio_manager.portfolio\
        .get_currency_portfolio(currency)
    current_market_quantity = exchange_manager.exchange_personal_data.portfolio_manager.portfolio\
        .get_currency_portfolio(market)

    market_quantity = current_market_quantity / mark_price if mark_price else 0

    symbol_market = exchange_manager.exchange.get_market_status(symbol, with_fixer=False)

    return current_symbol_holding, current_market_quantity, market_quantity, mark_price, symbol_market


def total_fees_from_order_dict(order_dict, currency):
    return get_fees_for_currency(order_dict[enums.ExchangeConstantsOrderColumns.FEE.value], currency)


def get_fees_for_currency(fee, currency):
    if fee and fee[enums.FeePropertyColumns.CURRENCY.value] == currency:
        return fee[enums.FeePropertyColumns.COST.value]
    return 0


def parse_order_status(raw_order):
    try:
        return enums.OrderStatus(raw_order[enums.ExchangeConstantsOrderColumns.STATUS.value])
    except KeyError:
        return KeyError("Could not parse new order status")


def parse_is_cancelled(raw_order):
    return parse_order_status(raw_order) in {enums.OrderStatus.CANCELED, enums.OrderStatus.CLOSED}



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.enums as channel_enums

import octobot_commons.channels_name as channels_name
import octobot_commons.constants as common_constants
import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.exchanges.exchanges as exchanges
import octobot_trading.modes.channel as modes_channel


class AbstractTradingModeProducer(modes_channel.ModeChannelProducer):
    def __init__(self, channel, config, trading_mode, exchange_manager):
        super().__init__(channel)
        # the trading mode instance logger
        self.logger = logging.get_logger(self.__class__.__name__)

        # the trading mode instance
        self.trading_mode = trading_mode

        # the global bot config
        self.config = config

        # the trading mode exchange manager
        self.exchange_manager = exchange_manager

        # shortcut
        self.exchange_name = self.exchange_manager.exchange_name

        # the final eval used by TradingModeConsumers, default value is INIT_EVAL_NOTE
        self.final_eval = common_constants.INIT_EVAL_NOTE

        # the producer state used by TradingModeConsumers
        self.state = None

        # the matrix consumer instance
        self.matrix_consumer = None

        # Define trading modes default consumer priority level
        self.priority_level: int = channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value

    # noinspection PyArgumentList
    async def start(self) -> None:
        """
        Start trading mode channels subscriptions
        """
        try:
            import octobot_evaluators.evaluators.channel as evaluators_channel
            import octobot_evaluators.enums as evaluators_enums
            matrix_id = exchanges.Exchanges.instance().get_exchange(self.exchange_manager.exchange_name,
                                                                    self.exchange_manager.id).matrix_id
            self.matrix_consumer = await evaluators_channel.get_chan(
                channels_name.OctoBotEvaluatorsChannelsName.MATRIX_CHANNEL.value, matrix_id).new_consumer(
                callback=self.matrix_callback,
                priority_level=self.priority_level,
                matrix_id=exchanges.Exchanges.instance().get_exchange(self.exchange_manager.exchange_name,
                                                                      self.exchange_manager.id).matrix_id,
                cryptocurrency=self.trading_mode.cryptocurrency if self.trading_mode.cryptocurrency \
                    else common_constants.CONFIG_WILDCARD,
                symbol=self.trading_mode.symbol if self.trading_mode.symbol else common_constants.CONFIG_WILDCARD,
                evaluator_type=evaluators_enums.EvaluatorMatrixTypes.STRATEGIES.value,
                exchange_name=self.exchange_name,
                time_frame=self.trading_mode.time_frame if self.trading_mode.time_frame
                else common_constants.CONFIG_WILDCARD)
        except (KeyError, ImportError):
            self.logger.error(f"Can't connect matrix channel on {self.exchange_name}")

    async def stop(self) -> None:
        """
        Stop trading mode channels subscriptions
        """
        await super().stop()
        if self.exchange_manager is not None:
            try:
                import octobot_evaluators.evaluators.channel as evaluators_channel
                await evaluators_channel.get_chan(channels_name.OctoBotEvaluatorsChannelsName.MATRIX_CHANNEL.value,
                                                  exchanges.Exchanges.instance().get_exchange(
                                                      self.exchange_manager.exchange_name,
                                                      self.exchange_manager.id).matrix_id
                                                  ).remove_consumer(self.matrix_consumer)
            except (KeyError, ImportError):
                self.logger.error(f"Can't unregister matrix channel on {self.exchange_name}")
        self.flush()

    def flush(self) -> None:
        """
        Flush all instance objects reference
        """
        self.trading_mode = None
        self.exchange_manager = None
        self.matrix_consumer = None

    async def matrix_callback(self, matrix_id, evaluator_name, evaluator_type,
                              eval_note, eval_note_type, exchange_name, cryptocurrency, symbol, time_frame) -> None:
        """
        Called when a strategy updates the matrix
        :param matrix_id: the matrix_id
        :param evaluator_name: the evaluator name, should be the strategy name
        :param evaluator_type: the evaluator type, should be EvaluatorMatrixTypes.STRATEGIES.value
        :param eval_note: the eval note, should be the strategy eval note
        :param eval_note_type: the eval note type
        :param exchange_name: the exchange name
        :param cryptocurrency: the cryptocurrency
        :param symbol: the symbol
        :param time_frame: the time frame
        """
        await self.finalize(exchange_name=exchange_name, matrix_id=matrix_id, cryptocurrency=cryptocurrency,
                            symbol=symbol, time_frame=time_frame)

    async def finalize(self, exchange_name: str,
                       matrix_id: str,
                       cryptocurrency: str = None,
                       symbol: str = None,
                       time_frame=None) -> None:
        """
        Finalize evaluation
        """
        if exchange_name != self.exchange_name or not self.exchange_manager.trader.is_enabled:
            # Do nothing if not its exchange
            return

        try:
            await self.set_final_eval(matrix_id=matrix_id,
                                      cryptocurrency=cryptocurrency,
                                      symbol=symbol,
                                      time_frame=time_frame)
        except Exception as e:
            self.logger.exception(e, True, f"Error when finalizing: {e}")

    async def set_final_eval(self, matrix_id: str, cryptocurrency: str, symbol: str, time_frame) -> None:
        """
        Called to calculate the final note or state => when any notification appears
        """
        raise NotImplementedError("set_final_eval not implemented")

    async def submit_trading_evaluation(self, cryptocurrency, symbol, time_frame,
                                        final_note=common_constants.INIT_EVAL_NOTE,
                                        state=enums.EvaluatorStates.NEUTRAL,
                                        data=None) -> None:
        await self.send(trading_mode_name=self.trading_mode.get_name(),
                        cryptocurrency=cryptocurrency,
                        symbol=symbol,
                        time_frame=time_frame,
                        final_note=final_note,
                        state=state.value,
                        data=data if data is not None else {})

    @classmethod
    def get_should_cancel_loaded_orders(cls) -> bool:
        """
        Called by cancel_symbol_open_orders => return true if OctoBot should cancel all orders for a symbol including
        orders already existing when OctoBot started up
        """
        raise NotImplementedError("get_should_cancel_loaded_orders not implemented")

    async def cancel_symbol_open_orders(self, symbol) -> bool:
        """
        Cancel all trader open orders
        """
        cancel_loaded_orders = self.get_should_cancel_loaded_orders()

        if self.exchange_manager.trader.is_enabled:
            return await self.exchange_manager.trader.cancel_open_orders(symbol, cancel_loaded_orders)
        return True



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_trading.enums import TradeOrderSide
from octobot_trading.personal_data.orders.types import BuyLimitOrder, SellLimitOrder, SellMarketOrder, BuyMarketOrder, \
    StopLossOrder, StopLossLimitOrder, TakeProfitOrder, TakeProfitLimitOrder, TrailingStopOrder, TrailingStopLimitOrder
from octobot_trading.personal_data.orders import Order


@pytest.fixture()
async def order(trader):
    config, trader_inst, exchange_manager = trader
    return config, trader_inst, exchange_manager, Order(trader_inst)


@pytest.fixture()
async def order_simulator(trader_simulator):
    config, trader_inst, exchange_manager = trader_simulator
    return config, trader_inst, exchange_manager, Order(trader_inst)


@pytest.fixture()
def buy_limit_order(event_loop, simulated_trader):
    _, _, trader_instance = simulated_trader
    return BuyLimitOrder(trader_instance)


@pytest.fixture()
def sell_limit_order(event_loop, simulated_trader):
    _, _, trader_instance = simulated_trader
    return SellLimitOrder(trader_instance)


@pytest.fixture()
def buy_market_order(event_loop, simulated_trader):
    _, _, trader_instance = simulated_trader
    return BuyMarketOrder(trader_instance)


@pytest.fixture()
def sell_market_order(event_loop, simulated_trader):
    _, _, trader_instance = simulated_trader
    return SellMarketOrder(trader_instance)


@pytest.fixture()
def stop_loss_sell_order(event_loop, simulated_trader):
    _, _, trader_instance = simulated_trader
    return StopLossOrder(trader_instance, side=TradeOrderSide.SELL)


@pytest.fixture()
def stop_loss_buy_order(event_loop, simulated_trader):
    _, _, trader_instance = simulated_trader
    return StopLossOrder(trader_instance, side=TradeOrderSide.BUY)


@pytest.fixture()
def stop_loss_limit_order(event_loop, simulated_trader):
    _, exchange_manager, trader_instance = simulated_trader
    return StopLossLimitOrder(trader_instance)


@pytest.fixture()
def take_profit_sell_order(event_loop, simulated_trader):
    _, _, trader_instance = simulated_trader
    return TakeProfitOrder(trader_instance, side=TradeOrderSide.SELL)


@pytest.fixture()
def take_profit_buy_order(event_loop, simulated_trader):
    _, _, trader_instance = simulated_trader
    return TakeProfitOrder(trader_instance, side=TradeOrderSide.BUY)


@pytest.fixture()
def take_profit_limit_order(event_loop, simulated_trader):
    _, exchange_manager, trader_instance = simulated_trader
    return TakeProfitLimitOrder(trader_instance)


@pytest.fixture()
def trailing_stop_order(event_loop, simulated_trader):
    _, _, trader_instance = simulated_trader
    return TrailingStopOrder(trader_instance)


@pytest.fixture()
def trailing_stop_limit_order(event_loop, simulated_trader):
    _, exchange_manager, trader_instance = simulated_trader
    return TrailingStopLimitOrder(trader_instance)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums


class FutureContract:
    def __init__(self, pair):
        self.pair = pair

        self.contract_type = enums.FutureContractType.INVERSE_PERPETUAL
        self.margin_type = enums.MarginType.ISOLATE
        self.expiration_timestamp = 0

        self.minimum_tick_size = 0.5
        self.contract_size = 1
        self.maximum_leverage = 1
        self.current_leverage = 1

    def is_inverse_contract(self):
        """
        :return: True if the contract is an inverse contract
        """
        return self.contract_type in [enums.FutureContractType.INVERSE_EXPIRABLE,
                                      enums.FutureContractType.INVERSE_PERPETUAL]

    def is_isolated(self):
        """
        :return: True if the contract use isolation margin
        """
        return self.margin_type is enums.MarginType.ISOLATE



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.positions cimport position
from octobot_trading.personal_data.positions.position cimport (
    Position,
    ShortPosition,
    LongPosition,
)

from octobot_trading.personal_data.positions cimport channel
from octobot_trading.personal_data.positions.channel cimport (
    PositionsProducer,
    PositionsChannel,
    PositionsUpdater,
    PositionsUpdaterSimulator,
)

from octobot_trading.personal_data.positions cimport positions_manager
from octobot_trading.personal_data.positions.positions_manager cimport (
    PositionsManager,
)

from octobot_trading.personal_data.positions cimport contracts
from octobot_trading.personal_data.positions.contracts cimport (
    FutureContract,
)

__all__ = [
    "PositionsProducer",
    "PositionsChannel",
    "PositionsUpdaterSimulator",
    "Position",
    "ShortPosition",
    "LongPosition",
    "PositionsUpdater",
    "PositionsManager",
    "FutureContract",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import pytest

from tests.personal_data.portfolios import update_portfolio_balance
from tests.test_utils.random_numbers import random_quantity, random_price

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_get_current_crypto_currencies_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    assert portfolio_value_holder.get_current_crypto_currencies_values() == {'BTC': 1, 'USDT': 0}
    update_portfolio_balance({
        'BTC': {'available': random_quantity(), 'total': random_quantity()},
        'ETH': {'available': random_quantity(), 'total': random_quantity()},
        'XRP': {'available': random_quantity(), 'total': random_quantity()},
        'NANO': {'available': random_quantity(), 'total': random_quantity()},
        'XLM': {'available': random_quantity(), 'total': random_quantity()},
        'USDT': {'available': random_quantity(), 'total': random_quantity()}
    }, exchange_manager)
    await portfolio_manager.handle_balance_updated()

    assert portfolio_value_holder.get_current_crypto_currencies_values() == {
        'BTC': 1,
        'ETH': 0,
        'XRP': 0,
        'NANO': 0,
        'XLM': 0,
        'USDT': 0
    }

    exchange_manager.client_symbols.append("XLM/BTC")
    exchange_manager.client_symbols.append("XRP/BTC")
    if not os.getenv('CYTHON_IGNORE'):
        portfolio_value_holder.missing_currency_data_in_exchange.remove("XRP")
        await portfolio_manager.handle_mark_price_update("XRP/BTC", 0.005)
        exchange_manager.client_symbols.append("NANO/BTC")
        portfolio_value_holder.missing_currency_data_in_exchange.remove("NANO")
        await portfolio_manager.handle_mark_price_update("NANO/BTC", 0.05)
        exchange_manager.client_symbols.append("BTC/USDT")

        assert portfolio_value_holder.get_current_crypto_currencies_values() == {
            'BTC': 1,
            'ETH': 0,
            'XRP': 0.005,
            'NANO': 0.05,
            'XLM': 0,
            'USDT': 0
        }
        xlm_btc_price = random_price(max_value=0.05)
        portfolio_value_holder.missing_currency_data_in_exchange.remove("XLM")
        await portfolio_manager.handle_mark_price_update("XLM/BTC", xlm_btc_price)
        assert portfolio_value_holder.get_current_crypto_currencies_values() == {
            'BTC': 1,
            'ETH': 0,
            'XRP': 0.005,
            'NANO': 0.05,
            'XLM': xlm_btc_price,
            'USDT': 0
        }
        usdt_btc_price = random_price(max_value=0.01)
        portfolio_value_holder.missing_currency_data_in_exchange.remove("USDT")
        await portfolio_manager.handle_mark_price_update("BTC/USDT", usdt_btc_price)
        assert portfolio_value_holder.get_current_crypto_currencies_values() == {
            'BTC': 1,
            'ETH': 0,
            'XRP': 0.005,
            'NANO': 0.05,
            'XLM': xlm_btc_price,
            'USDT': 1 / usdt_btc_price
        }
        eth_btc_price = random_price(max_value=1)
        exchange_manager.client_symbols.append("ETH/BTC")
        portfolio_value_holder.missing_currency_data_in_exchange.remove("ETH")
        await portfolio_manager.handle_mark_price_update("ETH/BTC", eth_btc_price)
        assert portfolio_value_holder.get_current_crypto_currencies_values() == {
            'BTC': 1,
            'ETH': eth_btc_price,
            'XRP': 0.005,
            'NANO': 0.05,
            'XLM': xlm_btc_price,
            'USDT': 1 / usdt_btc_price
        }


async def test_get_current_holdings_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    exchange_manager.client_symbols.append("ETH/BTC")
    update_portfolio_balance({
        'BTC': {'available': 10, 'total': 10},
        'ETH': {'available': 100, 'total': 100},
        'XRP': {'available': 10000, 'total': 10000},
        'USDT': {'available': 1000, 'total': 1000}
    }, exchange_manager)
    await portfolio_manager.handle_balance_updated()
    assert portfolio_value_holder.get_current_holdings_values() == {
        'BTC': 10,
        'ETH': 0,
        'XRP': 0,
        'USDT': 0
    }
    await portfolio_manager.handle_mark_price_update("ETH/BTC", 50)
    assert portfolio_value_holder.get_current_holdings_values() == {
        'BTC': 10,
        'ETH': 5000,
        'XRP': 0,
        'USDT': 0
    }
    await portfolio_manager.handle_mark_price_update("XRP/USDT", 2.5)
    assert portfolio_value_holder.get_current_holdings_values() == {
        'BTC': 10,
        'ETH': 5000,
        'XRP': 0,
        'USDT': 0
    }
    await portfolio_manager.handle_mark_price_update("XRP/BTC", 0.00001)
    assert portfolio_value_holder.get_current_holdings_values() == {
        'BTC': 10,
        'ETH': 5000,
        'XRP': 0,
        'USDT': 0
    }
    if not os.getenv('CYTHON_IGNORE'):
        exchange_manager.client_symbols.append("XRP/BTC")
        portfolio_value_holder.missing_currency_data_in_exchange.remove("XRP")
        await portfolio_manager.handle_mark_price_update("XRP/BTC", 0.00001)
        assert portfolio_value_holder.get_current_holdings_values() == {
            'BTC': 10,
            'ETH': 5000,
            'XRP': 0.1,
            'USDT': 0
        }
        exchange_manager.client_symbols.append("BTC/USDT")
        portfolio_value_holder.missing_currency_data_in_exchange.remove("USDT")
        await portfolio_manager.handle_mark_price_update("BTC/USDT", 5000)
        assert portfolio_value_holder.get_current_holdings_values() == {
            'BTC': 10,
            'ETH': 5000,
            'XRP': 0.1,
            'USDT': 0.2
        }


async def test_get_origin_portfolio_current_value(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    await portfolio_manager.handle_profitability_recalculation(True)
    assert portfolio_value_holder.get_origin_portfolio_current_value() == 10


async def test_get_origin_portfolio_current_value_with_different_reference_market(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    portfolio_manager.reference_market = "USDT"
    await portfolio_manager.handle_profitability_recalculation(True)
    assert portfolio_value_holder.get_origin_portfolio_current_value() == 1000


async def test_update_origin_crypto_currencies_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    assert portfolio_value_holder.update_origin_crypto_currencies_values("ETH/BTC", 0.1) is True
    assert portfolio_value_holder.origin_crypto_currencies_values["ETH"] == 0.1
    assert portfolio_value_holder.last_prices_by_trading_pair["ETH/BTC"] == 0.1

    assert portfolio_value_holder.update_origin_crypto_currencies_values("BTC/USDT", 100) is True
    assert portfolio_value_holder.origin_crypto_currencies_values["USDT"] == 1 / 100
    assert portfolio_value_holder.last_prices_by_trading_pair["BTC/USDT"] == 100



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.logging as logging

import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.constants
import octobot_trading.personal_data.portfolios.portfolio_manager as portfolio_manager
import octobot_trading.personal_data.positions.positions_manager as positions_manager
import octobot_trading.personal_data.orders.orders_manager as orders_manager
import octobot_trading.personal_data.trades.trades_manager as trades_manager
import octobot_trading.util as util


class ExchangePersonalData(util.Initializable):
    # note: symbol keys are without /
    def __init__(self, exchange_manager):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.exchange_manager = exchange_manager
        self.config = exchange_manager.config

        self.trader = None
        self.exchange = None
        self.portfolio_manager = None
        self.trades_manager = None
        self.orders_manager = None
        self.positions_manager = None

    async def initialize_impl(self):
        self.trader = self.exchange_manager.trader
        self.exchange = self.exchange_manager.exchange
        if self.trader.is_enabled:
            try:
                self.portfolio_manager = portfolio_manager.PortfolioManager(self.config, self.trader,
                                                                            self.exchange_manager)
                self.trades_manager = trades_manager.TradesManager(self.trader)
                self.orders_manager = orders_manager.OrdersManager(self.trader)
                self.positions_manager = positions_manager.PositionsManager(self.trader)
                await self.portfolio_manager.initialize()
                await self.trades_manager.initialize()
                await self.orders_manager.initialize()
                await self.positions_manager.initialize()
            except Exception as e:
                self.logger.exception(e, True, f"Error when initializing : {e}. "
                                               f"{self.exchange.name} personal data disabled.")

    # updates
    async def handle_portfolio_update(self, balance, should_notify: bool = True, is_diff_update=False) -> bool:
        try:
            changed: bool = self.portfolio_manager.handle_balance_update(balance, is_diff_update=is_diff_update)
            if should_notify:
                await exchange_channel.get_chan(octobot_trading.constants.BALANCE_CHANNEL,
                                         self.exchange_manager.id).get_internal_producer().send(balance)
            return changed
        except AttributeError as e:
            self.logger.exception(e, True, f"Failed to update balance : {e}")
            return False

    async def handle_portfolio_update_from_order(self, order, should_notify: bool = True) -> bool:
        try:
            changed: bool = await self.portfolio_manager.handle_balance_update_from_order(order)
            if should_notify:
                await exchange_channel.get_chan(octobot_trading.constants.BALANCE_CHANNEL, self.exchange_manager.id). \
                    get_internal_producer().send(self.portfolio_manager.portfolio.portfolio)
            return changed
        except AttributeError as e:
            self.logger.exception(e, True, f"Failed to update balance : {e}")
            return False

    async def handle_portfolio_profitability_update(self, balance, mark_price, symbol, should_notify: bool = True):
        try:
            portfolio_profitability = self.portfolio_manager.portfolio_profitability

            if balance is not None:
                await self.portfolio_manager.handle_balance_updated()

            if mark_price is not None and symbol is not None:
                await self.portfolio_manager.handle_mark_price_update(symbol=symbol, mark_price=mark_price)

            if should_notify:
                await exchange_channel.get_chan(octobot_trading.constants.BALANCE_PROFITABILITY_CHANNEL,
                                         self.exchange_manager.id).get_internal_producer() \
                    .send(profitability=portfolio_profitability.profitability,
                          profitability_percent=portfolio_profitability.profitability_percent,
                          market_profitability_percent=portfolio_profitability.market_profitability_percent,
                          initial_portfolio_current_profitability=portfolio_profitability.initial_portfolio_current_profitability)
        except Exception as e:
            self.logger.exception(e, True, f"Failed to update portfolio profitability : {e}")

    async def handle_order_update_from_raw(self, order_id, raw_order,
                                           is_new_order: bool = False,
                                           should_notify: bool = True) -> bool:
        # Orders can sometimes be out of sync between different exchange endpoints (ex: binance order API vs
        # open_orders API which is slower).
        # Always check if this order has not already been closed previously (most likely during the last
        # seconds/minutes)
        if self._is_out_of_sync_order(order_id):
            self.logger.debug(f"Ignored update for order with {order_id}: this order has already been closed "
                              f"(received raw order: {raw_order})")
        else:
            try:
                changed: bool = await self.orders_manager.upsert_order_from_raw(order_id, raw_order)

                if changed:
                    updated_order = self.orders_manager.get_order(order_id)
                    asyncio.create_task(updated_order.state.on_order_refresh_successful())

                    if should_notify:
                        await self.handle_order_update_notification(updated_order, is_new_order)

                return changed
            except KeyError as ke:
                self.logger.debug(f"Failed to update order : Order was not found ({ke})")
            except Exception as e:
                self.logger.exception(e, True, f"Failed to update order : {e}")
        return False

    def _is_out_of_sync_order(self, order_id) -> bool:
        return self.trades_manager.has_closing_trade_with_order_id(order_id)

    async def handle_order_instance_update(self, order, is_new_order: bool = False, should_notify: bool = True):
        try:
            changed: bool = self.orders_manager.upsert_order_instance(order)

            if changed:
                asyncio.create_task(order.state.on_order_refresh_successful())

                if should_notify:
                    await self.handle_order_update_notification(order, is_new_order)

            return changed
        except Exception as e:
            self.logger.exception(e, True, f"Failed to update order instance : {e}")
            return False

    async def handle_order_update_notification(self, order, is_new_order):
        """
        Notify Orders channel for Order update
        :param order: the updated order
        :param is_new_order: True if the order was created during update
        """
        try:
            await exchange_channel.get_chan(octobot_trading.constants.ORDERS_CHANNEL,
                                     self.exchange_manager.id).get_internal_producer() \
                .send(cryptocurrency=self.exchange_manager.exchange.get_pair_cryptocurrency(order.symbol),
                      symbol=order.symbol,
                      order=order.to_dict(),
                      is_from_bot=order.is_from_this_octobot,
                      is_new=is_new_order,
                      is_closed=order.is_closed())
        except ValueError as e:
            self.logger.error(f"Failed to send order update notification : {e}")

    async def handle_closed_order_update(self, order_id, raw_order) -> bool:
        """
        Handle closed order creation or update
        :param order_id: the closed order id
        :param raw_order: the closed order dict
        :return: True if the closed order has been created or updated
        """
        try:
            return await self.orders_manager.upsert_order_close_from_raw(order_id, raw_order) is not None
        except Exception as e:
            self.logger.exception(e, True, f"Failed to update order : {e}")
            return False

    async def handle_trade_update(self, symbol, trade_id, trade, should_notify: bool = True):
        try:
            changed: bool = self.trades_manager.upsert_trade(trade_id, trade)
            if should_notify:
                await exchange_channel.get_chan(octobot_trading.constants.TRADES_CHANNEL,
                                         self.exchange_manager.id).get_internal_producer() \
                    .send(cryptocurrency=self.exchange_manager.exchange.get_pair_cryptocurrency(symbol),
                          symbol=symbol,
                          trade=trade.to_dict(),
                          old_trade=False)
            return changed
        except Exception as e:
            self.logger.exception(e, True, f"Failed to update trade : {e}")
            return False

    async def handle_trade_instance_update(self, trade, should_notify: bool = True):
        try:
            changed: bool = self.trades_manager.upsert_trade_instance(trade)
            if should_notify:
                await exchange_channel.get_chan(octobot_trading.constants.TRADES_CHANNEL,
                               self.exchange_manager.id).get_internal_producer() \
                    .send(cryptocurrency=self.exchange_manager.exchange.get_pair_cryptocurrency(trade.symbol),
                          symbol=trade.symbol,
                          trade=trade.to_dict(),
                          old_trade=False)
            return changed
        except Exception as e:
            self.logger.exception(e, True, f"Failed to update trade instance : {e}")
            return False

    async def handle_position_update(self, symbol, position_id, position, should_notify: bool = True):
        try:
            changed: bool = self.positions_manager.upsert_position(position_id, position)
            if should_notify:
                position_instance = self.positions_manager[position_id]
                await exchange_channel.get_chan(octobot_trading.constants.POSITIONS_CHANNEL,
                                         self.exchange_manager.id).get_internal_producer() \
                    .send(cryptocurrency=self.exchange_manager.exchange.get_pair_cryptocurrency(symbol),
                          symbol=symbol,
                          position=position_instance.to_dict(),
                          is_closed=False,
                          is_updated=changed,
                          is_from_bot=True,
                          is_liquidated=position_instance.is_liquidated())
            return changed
        except Exception as e:
            self.logger.exception(e, True, f"Failed to update position : {e}")
            return False

    async def handle_position_instance_update(self, position, should_notify: bool = True):
        try:
            changed: bool = self.positions_manager.upsert_position_instance(position)
            if should_notify:
                await exchange_channel.get_chan(octobot_trading.constants.POSITIONS_CHANNEL,
                                         self.exchange_manager.id).get_internal_producer() \
                    .send(cryptocurrency=self.exchange_manager.exchange.get_pair_cryptocurrency(position.symbol),
                          symbol=position.symbol,
                          position=position,
                          is_closed=False,
                          is_updated=changed,
                          is_from_bot=True,
                          is_liquidated=position.is_liquidated())
            return changed
        except Exception as e:
            self.logger.exception(e, True, f"Failed to update position instance : {e}")
            return False

    def get_order_portfolio(self, order):
        return order.linked_portfolio if order.linked_portfolio is not None else self.portfolio_manager.portfolio

    def clear(self):
        if self.portfolio_manager is not None:
            self.portfolio_manager.clear()
        if self.orders_manager is not None:
            self.orders_manager.clear()
        if self.trades_manager is not None:
            self.trades_manager.clear()



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import typing

import time

import octobot_trading.errors as errors
import octobot_commons.constants as common_constants

import octobot_trading.exchange_data.funding.channel.funding as funding_channel
import octobot_trading.constants as constants
import octobot_trading.enums as enums


class FundingUpdater(funding_channel.FundingProducer):
    """
    The Funding Update fetch the exchange funding rate and send it to the Funding Channel
    """

    """
    The updater related channel name
    """
    CHANNEL_NAME = constants.FUNDING_CHANNEL

    """
    The default funding update refresh times in seconds
    """
    FUNDING_REFRESH_TIME = 2 * common_constants.HOURS_TO_SECONDS
    FUNDING_REFRESH_TIME_MIN = 0.2 * common_constants.HOURS_TO_SECONDS
    FUNDING_REFRESH_TIME_MAX = 8 * common_constants.HOURS_TO_SECONDS

    async def start(self) -> None:
        """
        Starts the funding fetching process
        """
        if not self._should_run():
            return
        if self.channel.is_paused:
            await self.pause()
        else:
            await self.start_update_loop()

    async def start_update_loop(self):
        while not self.should_stop and not self.channel.is_paused:
            next_funding_time, sleep_time = await self.before_update()
            try:
                for (
                    pair
                ) in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                    next_funding_time_candidate = await self.fetch_symbol_funding_rate(
                        pair
                    )
                    if next_funding_time_candidate is not None:
                        next_funding_time = next_funding_time_candidate
            except (errors.NotSupported, NotImplementedError):
                self.logger.warning(
                    f"{self.channel.exchange_manager.exchange_name} is not supporting updates"
                )
                await self.pause()
            finally:
                if next_funding_time:
                    should_sleep_time = next_funding_time - time.time()
                    sleep_time = (
                        should_sleep_time
                        if should_sleep_time < self.FUNDING_REFRESH_TIME_MAX
                        else self.FUNDING_REFRESH_TIME
                    )
                await asyncio.sleep(sleep_time)

    async def before_update(self) -> (int, int):
        """
        Called to initialize funding update
        :return: the next funding time and the sleep time
        """
        return None, self.FUNDING_REFRESH_TIME_MIN

    async def fetch_symbol_funding_rate(self, symbol: str) -> typing.Optional[int]:
        """
        Fetch funding rate from exchange
        :param symbol: the funding symbol to fetch
        :return: the next funding time
        """
        try:
            funding: dict = await self.channel.exchange_manager.exchange.get_funding_rate(
                symbol
            )

            if funding:
                next_funding_time = funding[
                    enums.ExchangeConstantsFundingColumns.NEXT_FUNDING_TIME.value
                ]
                await self.push(
                    symbol=symbol,
                    funding_rate=funding[
                        enums.ExchangeConstantsFundingColumns.FUNDING_RATE.value
                    ],
                    next_funding_time=next_funding_time,
                    timestamp=funding[
                        enums.ExchangeConstantsFundingColumns.LAST_FUNDING_TIME.value
                    ],
                )
                return next_funding_time
        except (errors.NotSupported, NotImplementedError) as ne:
            raise ne
        except Exception as e:
            self.logger.exception(e, True, f"Fail to update funding rate : {e}")
        return None

    def _should_run(self) -> bool:
        """
        Check if the funding update should run
        :return: the check result
        """
        if not self.channel.exchange_manager.is_future:
            return False
        return (
            not self.channel.exchange_manager.exchange.FUNDING_WITH_MARK_PRICE
            and not self.channel.exchange_manager.exchange.FUNDING_IN_TICKER
        )

    async def resume(self) -> None:
        """
        Resume updater process
        """
        if not self._should_run():
            return
        await super().resume()
        if not self.is_running:
            await self.run()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.errors as errors
import octobot_trading.exchanges as exchanges


async def create_exchanges(exchange_manager):
    if exchange_manager.is_sandboxed:
        exchange_manager.logger.info(f"Using sandbox exchange for {exchange_manager.exchange_name}")

    if not exchange_manager.is_backtesting:
        # real : create a rest or websocket exchange instance
        await create_real_exchange(exchange_manager)
        exchange_manager.load_constants()
        await initialize_real_exchange(exchange_manager)
    else:
        # simulated : create exchange simulator instance
        await create_simulated_exchange(exchange_manager)

    if not exchange_manager.exchange_only:
        # create exchange producers if necessary
        await exchanges.create_exchange_producers(exchange_manager)

    if exchange_manager.is_backtesting:
        await init_simulated_exchange(exchange_manager)

    exchange_manager.exchange_name = exchange_manager.exchange.name
    exchange_manager.is_ready = True


async def create_real_exchange(exchange_manager) -> None:
    """
    Create and initialize real REST exchange
    :param exchange_manager: the related exchange manager
    """
    await _create_rest_exchange(exchange_manager)

    try:
        await exchange_manager.exchange.initialize()
    except errors.AuthenticationError:
        exchange_manager.logger.error("Authentication error, retrying without authentication...")
        exchange_manager.without_auth = True
        await create_real_exchange(exchange_manager)
        return


async def initialize_real_exchange(exchange_manager):
    if not exchange_manager.exchange_only:
        await exchanges.create_exchange_channels(exchange_manager)

    # create Websocket exchange if possible
    if not exchange_manager.rest_only:
        # search for websocket
        if exchanges.check_web_socket_config(exchange_manager.config, exchange_manager.exchange.name):
            await exchanges.search_and_create_websocket(exchange_manager)


async def _create_rest_exchange(exchange_manager) -> None:
    """
    create REST based on ccxt exchange
    :param exchange_manager: the related exchange manager
    """
    if exchange_manager.is_future and not exchange_manager.is_spot_only:
        await _search_and_create_future_exchange(exchange_manager)
    elif exchange_manager.is_margin and not exchange_manager.is_spot_only:
        await _search_and_create_margin_exchange(exchange_manager)
    else:
        await _search_and_create_spot_exchange(exchange_manager)

    if not exchange_manager.exchange:
        raise Exception("Can't create an exchange instance that match the exchange configuration")


async def create_simulated_exchange(exchange_manager):
    if exchange_manager.is_future and not exchange_manager.is_spot_only:
        exchange_manager.exchange = exchanges.FutureExchangeSimulator(exchange_manager.config, exchange_manager,
                                                                      exchange_manager.backtesting)
    elif exchange_manager.is_margin and not exchange_manager.is_spot_only:
        exchange_manager.exchange = exchanges.MarginExchangeSimulator(exchange_manager.config, exchange_manager,
                                                                      exchange_manager.backtesting)
    else:
        exchange_manager.exchange = exchanges.SpotExchangeSimulator(exchange_manager.config, exchange_manager,
                                                                    exchange_manager.backtesting)

    await exchange_manager.exchange.initialize()
    _initialize_simulator_time_frames(exchange_manager)
    exchange_manager.exchange_config.set_config_time_frame()
    exchange_manager.exchange_config.set_config_traded_pairs()
    await exchanges.create_exchange_channels(exchange_manager)


async def init_simulated_exchange(exchange_manager):
    await exchange_manager.exchange.create_backtesting_exchange_producers()


async def _search_and_create_spot_exchange(exchange_manager) -> None:
    """
    Create a spot exchange if a SpotExchange matching class is found
    :param exchange_manager: the related exchange manager
    """
    spot_exchange_class = exchanges.get_spot_exchange_class(exchange_manager.exchange_name,
                                                            exchange_manager.tentacles_setup_config)
    if spot_exchange_class:
        exchange_manager.exchange = spot_exchange_class(config=exchange_manager.config,
                                                        exchange_manager=exchange_manager)


async def _search_and_create_margin_exchange(exchange_manager) -> None:
    """
    Create a margin exchange if a MarginExchange matching class is found
    :param exchange_manager: the related exchange manager
    """
    margin_exchange_class = exchanges.get_margin_exchange_class(exchange_manager.exchange_name,
                                                                exchange_manager.tentacles_setup_config)
    if margin_exchange_class:
        exchange_manager.exchange = margin_exchange_class(config=exchange_manager.config,
                                                          exchange_manager=exchange_manager)


async def _search_and_create_future_exchange(exchange_manager) -> None:
    """
    Create a future exchange if a FutureExchange matching class is found
    :param exchange_manager: the related exchange manager
    """
    future_exchange_class = exchanges.get_future_exchange_class(exchange_manager.exchange_name,
                                                                exchange_manager.tentacles_setup_config)
    if future_exchange_class:
        exchange_manager.exchange = future_exchange_class(config=exchange_manager.config,
                                                          exchange_manager=exchange_manager)


def _initialize_simulator_time_frames(exchange_manager):
    """
    Initialize simulator client time frames
    :param exchange_manager: the related exchange manager
    """
    exchange_manager.client_time_frames = exchange_manager.exchange.get_available_time_frames()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import typing
import contextlib

import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.personal_data.orders.states as orders_states
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.util as util


class Order(util.Initializable):
    """
    Order class will represent an open order in the specified exchange
    In simulation it will also define rules to be filled / canceled
    It is also use to store creation & fill values of the order
    """
    CHECK_ORDER_STATUS_AFTER_INIT_DELAY = 2

    def __init__(self, trader, side=None):
        super().__init__()
        self.trader = trader
        self.exchange_manager = trader.exchange_manager
        self.status = enums.OrderStatus.OPEN
        self.creation_time = self.exchange_manager.exchange.get_exchange_current_time()
        self.executed_time = 0
        self.lock = asyncio.Lock()
        self.linked_orders = []

        self.is_synchronized_with_exchange = False
        self.is_from_this_octobot = True
        self.order_id = trader.parse_order_id(None)
        self.simulated = trader.simulate

        self.logger_name = None
        self.symbol = None
        self.currency = None
        self.market = None
        self.taker_or_maker = None
        self.timestamp = 0
        self.origin_price = 0
        self.created_last_price = 0
        self.origin_quantity = 0
        self.origin_stop_price = 0
        self.order_type = None
        self.side = side
        self.filled_quantity = 0
        self.linked_portfolio = None
        self.linked_to = None
        self.canceled_time = 0
        self.fee = None
        self.filled_price = 0
        self.order_profitability = 0
        self.total_cost = 0

        # raw exchange order type, used to create order dict
        self.exchange_order_type = None

        # order state is initialized in initialize_impl()
        self.state = None

    @classmethod
    def get_name(cls):
        return cls.__name__

    def get_logger_name(self):
        if self.logger_name is None:
            self.logger_name = f"{self.get_name()} | {self.order_id}"
        return self.logger_name

    def update(self, symbol, order_id="", status=enums.OrderStatus.OPEN,
               current_price=0.0, quantity=0.0, price=0.0, stop_price=0.0,
               quantity_filled=0.0, filled_price=0.0, average_price=0.0, fee=None, total_cost=0.0,
               timestamp=None, linked_to=None, linked_portfolio=None, order_type=None) -> bool:
        changed: bool = False

        if order_id and self.order_id != order_id:
            self.order_id = order_id

        if symbol and self.symbol != symbol:
            self.currency, self.market = self.exchange_manager.get_exchange_quote_and_base(symbol)
            self.symbol = symbol

        if status and self.status != status:
            self.status = status
            changed = True
        if not self.status:
            self.status = enums.OrderStatus.OPEN

        if timestamp and self.timestamp != timestamp:
            self.timestamp = timestamp
        if not self.timestamp:
            if not timestamp:
                self.creation_time = self.exchange_manager.exchange.get_exchange_current_time()
            else:
                # if we have a timestamp, it's a real trader => need to format timestamp if necessary
                self.creation_time = self.exchange_manager.exchange.get_uniform_timestamp(timestamp)
            self.timestamp = self.creation_time

        if status in {enums.OrderStatus.FILLED, enums.OrderStatus.CLOSED} and not self.executed_time:
            self.executed_time = self.timestamp

        if price and self.origin_price != price:
            self.origin_price = price
            changed = True

        if fee is not None and self.fee != fee:
            self.fee = fee

        if current_price and self.created_last_price != current_price:
            self.created_last_price = current_price
            changed = True

        if quantity and self.origin_quantity != quantity:
            self.origin_quantity = quantity
            changed = True

        if stop_price and self.origin_stop_price != stop_price:
            self.origin_stop_price = stop_price
            changed = True

        if total_cost and self.total_cost != total_cost:
            self.total_cost = total_cost

        if average_price:
            # try using average price first
            if self.filled_price != average_price:
                self.filled_price = average_price
        else:
            if filled_price and self.filled_price != filled_price:
                self.filled_price = filled_price

        if self.trader.simulate:
            if quantity and not self.filled_quantity:
                self.filled_quantity = quantity
                changed = True

            if quantity_filled and self.filled_quantity != quantity_filled:
                self.filled_quantity = quantity_filled
                changed = True
        else:
            if quantity_filled and self.filled_quantity != quantity_filled:
                self.filled_quantity = quantity_filled
                changed = True

        if linked_to:
            self.linked_to = linked_to

        if linked_portfolio:
            self.linked_portfolio = linked_portfolio

        if order_type:
            self.order_type = order_type
            if self.exchange_order_type is None:
                self.exchange_order_type = _get_trade_order_type(order_type)

        if not self.filled_price and self.filled_quantity and self.total_cost:
            self.filled_price = self.total_cost / self.filled_quantity
            if timestamp is not None:
                self.executed_time = self.exchange_manager.exchange.get_uniform_timestamp(timestamp)

        if self.taker_or_maker is None:
            self._update_taker_maker()
        return changed

    async def initialize_impl(self, **kwargs):
        """
        Initialize order status update tasks
        """
        await orders_states.create_order_state(self, **kwargs)
        if not self.is_closed():
            await self.update_order_status()

    async def update_order_status(self, force_refresh=False):
        """
        Update_order_status will define the rules for a simulated order to be filled / canceled
        """
        raise NotImplementedError("Update_order_status not implemented")

    def add_linked_order(self, order):
        self.linked_orders.append(order)

    def get_currency_and_market(self) -> (str, str):
        return self.currency, self.market

    def get_total_fees(self, currency):
        return order_util.get_fees_for_currency(self.fee, currency)

    def is_open(self):
        return self.state is None or self.state.is_open()

    def is_filled(self):
        return self.state.is_filled() or (self.state.is_closed() and self.status is enums.OrderStatus.FILLED)

    def is_cancelled(self):
        return self.state.is_canceled() or (self.state.is_closed() and self.status is enums.OrderStatus.CANCELED)

    def is_closed(self):
        return self.state.is_closed() if self.state is not None else self.status is enums.OrderStatus.CLOSED

    def is_refreshing(self):
        return self.state is not None and self.state.is_refreshing()

    @contextlib.contextmanager
    def order_state_creation(self):
        try:
            yield
        except errors.InvalidOrderState as exc:
            logging.get_logger(self.get_logger_name()).exception(exc, True, f"Error when creating order state: {exc}")

    async def on_open(self, force_open=False, is_from_exchange_data=False):
        with self.order_state_creation():
            self.state = orders_states.OpenOrderState(self, is_from_exchange_data=is_from_exchange_data)
            await self.state.initialize(forced=force_open)

    async def on_fill(self, force_fill=False, is_from_exchange_data=False):
        logging.get_logger(self.get_logger_name()).debug(f"on_fill triggered for {self}")
        if self.is_open() and not self.is_refreshing():
            with self.order_state_creation():
                self.state = orders_states.FillOrderState(self, is_from_exchange_data=is_from_exchange_data)
                await self.state.initialize(forced=force_fill)
        else:
            logging.get_logger(self.get_logger_name()).debug(f"Trying to fill a refreshing or previously filled "
                                                             f"or canceled order: "
                                                             f"ignored fill call for {self}")

    async def on_close(self, force_close=False, is_from_exchange_data=False):
        with self.order_state_creation():
            self.state = orders_states.CloseOrderState(self, is_from_exchange_data=is_from_exchange_data)
            await self.state.initialize(forced=force_close)

    async def on_cancel(self, is_from_exchange_data=False, force_cancel=False, ignored_order=None):
        with self.order_state_creation():
            self.state = orders_states.CancelOrderState(self, is_from_exchange_data=is_from_exchange_data)
            await self.state.initialize(forced=force_cancel, ignored_order=ignored_order)

    def on_fill_actions(self):
        """
        Perform on_fill actions
        """
        self.status = enums.OrderStatus.FILLED

    async def on_filled(self):
        """
        Filling complete callback
        """
        # nothing to do by default

    def get_computed_fee(self, forced_value=None):
        computed_fee = self.exchange_manager.exchange.get_trade_fee(self.symbol, self.order_type, self.filled_quantity,
                                                                    self.filled_price, self.taker_or_maker)
        return {
            enums.FeePropertyColumns.COST.value:
                forced_value if forced_value is not None else computed_fee[enums.FeePropertyColumns.COST.value],
            enums.FeePropertyColumns.CURRENCY.value: computed_fee[enums.FeePropertyColumns.CURRENCY.value],
        }

    def get_profitability(self):
        if self.filled_price != 0 and self.created_last_price != 0:
            if self.filled_price >= self.created_last_price:
                self.order_profitability = 1 - self.filled_price / self.created_last_price
                if self.side is enums.TradeOrderSide.SELL:
                    self.order_profitability *= -1
            else:
                self.order_profitability = 1 - self.created_last_price / self.filled_price
                if self.side is enums.TradeOrderSide.BUY:
                    self.order_profitability *= -1
        return self.order_profitability

    async def default_exchange_update_order_status(self):
        result = await self.exchange_manager.exchange.get_order(self.order_id, self.symbol)
        new_status = self.trader.parse_status(result)
        self.is_synchronized_with_exchange = True
        if new_status in {enums.OrderStatus.FILLED, enums.OrderStatus.CLOSED}:
            await self.on_fill()
        elif new_status is enums.OrderStatus.CANCELED:
            await self.trader.cancel_order(self)

    def generate_executed_time(self):
        return self.exchange_manager.exchange.get_exchange_current_time()

    def is_self_managed(self):
        return not self.exchange_manager.exchange.is_supported_order_type(self.order_type)

    def update_from_raw(self, raw_order):
        if self.side is None or self.order_type is None:
            try:
                self._update_type_from_raw(raw_order)
                if self.taker_or_maker is None:
                    self._update_taker_maker()
            except KeyError:
                logging.get_logger(self.__class__.__name__).warning("Failed to parse order side and type")

        filled_price = raw_order.get(enums.ExchangeConstantsOrderColumns.PRICE.value, 0.0)
        # set average price with real average price if available, use filled_price otherwise
        average_price = raw_order.get(enums.ExchangeConstantsOrderColumns.AVERAGE.value, 0.0) or filled_price

        return self.update(
            symbol=str(raw_order.get(enums.ExchangeConstantsOrderColumns.SYMBOL.value, None)),
            current_price=raw_order.get(enums.ExchangeConstantsOrderColumns.PRICE.value, 0.0),
            quantity=raw_order.get(enums.ExchangeConstantsOrderColumns.AMOUNT.value, 0.0),
            price=raw_order.get(enums.ExchangeConstantsOrderColumns.PRICE.value, 0.0),
            status=order_util.parse_order_status(raw_order),
            order_id=str(raw_order.get(enums.ExchangeConstantsOrderColumns.ID.value, None)),
            quantity_filled=raw_order.get(enums.ExchangeConstantsOrderColumns.FILLED.value, 0.0),
            filled_price=filled_price,
            average_price=average_price,
            total_cost=raw_order.get(enums.ExchangeConstantsOrderColumns.COST.value, 0.0),
            fee=raw_order.get(enums.ExchangeConstantsOrderColumns.FEE.value, None),
            timestamp=raw_order.get(enums.ExchangeConstantsOrderColumns.TIMESTAMP.value, None)
        )

    def consider_as_filled(self):
        self.status = enums.OrderStatus.FILLED
        if self.executed_time == 0:
            self.executed_time = self.timestamp
        if self.filled_quantity == 0:
            self.filled_quantity = self.origin_quantity
        if self.filled_price == 0:
            self.filled_price = self.origin_price

    def update_order_from_raw(self, raw_order):
        self.status = order_util.parse_order_status(raw_order)
        self.total_cost = raw_order[enums.ExchangeConstantsOrderColumns.COST.value]
        self.filled_quantity = raw_order[enums.ExchangeConstantsOrderColumns.FILLED.value]
        self.filled_price = raw_order[enums.ExchangeConstantsOrderColumns.PRICE.value]
        if not self.filled_price and self.filled_quantity:
            self.filled_price = self.total_cost / self.filled_quantity

        self._update_taker_maker()

        self.fee = raw_order[enums.ExchangeConstantsOrderColumns.FEE.value]

        self.executed_time = self.trader.exchange.get_uniform_timestamp(
            raw_order[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value])

    def _update_type_from_raw(self, raw_order):
        try:
            self.exchange_order_type = enums.TradeOrderType(raw_order[enums.ExchangeConstantsOrderColumns.TYPE.value])
        except ValueError:
            self.exchange_order_type = enums.TradeOrderType.UNKNOWN
        self.side, self.order_type = parse_order_type(raw_order)

    def _update_taker_maker(self):
        if self.order_type in [enums.TraderOrderType.SELL_MARKET,
                               enums.TraderOrderType.BUY_MARKET,
                               enums.TraderOrderType.STOP_LOSS]:
            # always true
            self.taker_or_maker = enums.ExchangeConstantsMarketPropertyColumns.TAKER.value
        else:
            # true 90% of the time: impossible to know for sure the reality
            # (should only be used for simulation anyway)
            self.taker_or_maker = enums.ExchangeConstantsMarketPropertyColumns.MAKER.value

    def to_dict(self):
        filled_price = self.filled_price if self.filled_price > 0 else self.origin_price
        return {
            enums.ExchangeConstantsOrderColumns.ID.value: self.order_id,
            enums.ExchangeConstantsOrderColumns.SYMBOL.value: self.symbol,
            enums.ExchangeConstantsOrderColumns.PRICE.value: filled_price,
            enums.ExchangeConstantsOrderColumns.STATUS.value: self.status.value,
            enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: self.timestamp,
            enums.ExchangeConstantsOrderColumns.TYPE.value: self.exchange_order_type.value
            if self.exchange_order_type else None,
            enums.ExchangeConstantsOrderColumns.SIDE.value: self.side.value,
            enums.ExchangeConstantsOrderColumns.AMOUNT.value: self.origin_quantity,
            enums.ExchangeConstantsOrderColumns.COST.value: self.total_cost,
            enums.ExchangeConstantsOrderColumns.FILLED.value: self.filled_quantity,
            enums.ExchangeConstantsOrderColumns.FEE.value: self.fee
        }

    def clear(self):
        self.state.clear()
        self.trader = None
        self.exchange_manager = None
        self.linked_to = None
        self.linked_portfolio = None
        self.linked_orders = []

    def is_cleared(self):
        return self.exchange_manager is None

    def to_string(self):
        return (f"{self.symbol} | "
                f"{self.order_type.name if self.order_type is not None else 'Unknown'} | "
                f"Price : {self.origin_price} | "
                f"Quantity : {self.origin_quantity} | "
                f"State : {self.state.state.value if self.state is not None else 'Unknown'} | "
                f"id : {self.order_id}")

    def __str__(self):
        return self.to_string()

    def is_to_be_maintained(self):
        return self.trader is not None


def parse_order_type(raw_order):
    try:
        side: enums.TradeOrderSide = enums.TradeOrderSide(raw_order[enums.ExchangeConstantsOrderColumns.SIDE.value])
        order_type: enums.TradeOrderType = enums.TradeOrderType.UNKNOWN
        parsed_order_type: enums.TraderOrderType = enums.TraderOrderType.UNKNOWN
        try:
            order_type = enums.TradeOrderType(raw_order[enums.ExchangeConstantsOrderColumns.TYPE.value])
        except ValueError as e:
            if raw_order[enums.ExchangeConstantsOrderColumns.TYPE.value] is None:
                # No order type info: use unknown order type
                return side, enums.TraderOrderType.UNKNOWN
            else:
                # Incompatible order type info: raise error
                raise e

        if order_type == enums.TradeOrderType.UNKNOWN:
            parsed_order_type = enums.TraderOrderType.UNKNOWN
        elif side == enums.TradeOrderSide.BUY:
            if order_type == enums.TradeOrderType.LIMIT or order_type == enums.TradeOrderType.LIMIT_MAKER:
                parsed_order_type = enums.TraderOrderType.BUY_LIMIT
            elif order_type == enums.TradeOrderType.MARKET:
                parsed_order_type = enums.TraderOrderType.BUY_MARKET
            else:
                parsed_order_type = _get_sell_and_buy_types(order_type)
        elif side == enums.TradeOrderSide.SELL:
            if order_type == enums.TradeOrderType.LIMIT or order_type == enums.TradeOrderType.LIMIT_MAKER:
                parsed_order_type = enums.TraderOrderType.SELL_LIMIT
            elif order_type == enums.TradeOrderType.MARKET:
                parsed_order_type = enums.TraderOrderType.SELL_MARKET
            else:
                parsed_order_type = _get_sell_and_buy_types(order_type)
        return side, parsed_order_type
    except (KeyError, ValueError):
        return None, None


def _get_trade_order_type(order_type: enums.TraderOrderType) -> typing.Optional[enums.TradeOrderType]:
    if order_type == enums.TraderOrderType.BUY_LIMIT or order_type == enums.TraderOrderType.SELL_LIMIT:
        return enums.TradeOrderType.LIMIT
    if order_type == enums.TraderOrderType.BUY_MARKET or order_type == enums.TraderOrderType.SELL_MARKET:
        return enums.TradeOrderType.MARKET
    elif order_type == enums.TraderOrderType.TAKE_PROFIT:
        return enums.TradeOrderType.TAKE_PROFIT
    elif order_type == enums.TraderOrderType.TAKE_PROFIT_LIMIT:
        return enums.TradeOrderType.TAKE_PROFIT_LIMIT
    elif order_type == enums.TraderOrderType.STOP_LOSS:
        return enums.TradeOrderType.STOP_LOSS
    elif order_type == enums.TraderOrderType.STOP_LOSS_LIMIT:
        return enums.TradeOrderType.STOP_LOSS_LIMIT
    elif order_type == enums.TraderOrderType.TRAILING_STOP:
        return enums.TradeOrderType.TRAILING_STOP
    elif order_type == enums.TraderOrderType.TRAILING_STOP_LIMIT:
        return enums.TradeOrderType.TRAILING_STOP_LIMIT
    return None


def _get_sell_and_buy_types(order_type) -> typing.Optional[enums.TraderOrderType]:
    if order_type == enums.TradeOrderType.STOP_LOSS:
        return enums.TraderOrderType.STOP_LOSS
    elif order_type == enums.TradeOrderType.STOP_LOSS_LIMIT:
        return enums.TraderOrderType.STOP_LOSS_LIMIT
    elif order_type == enums.TradeOrderType.TAKE_PROFIT:
        return enums.TraderOrderType.TAKE_PROFIT
    elif order_type == enums.TradeOrderType.TAKE_PROFIT_LIMIT:
        return enums.TraderOrderType.TAKE_PROFIT_LIMIT
    elif order_type == enums.TradeOrderType.TRAILING_STOP:
        return enums.TraderOrderType.TRAILING_STOP
    elif order_type == enums.TradeOrderType.TRAILING_STOP_LIMIT:
        return enums.TraderOrderType.TRAILING_STOP_LIMIT
    return None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy

import octobot_commons.logging as logging

import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.constants as constants
import octobot_trading.personal_data.positions.channel.positions_updater as positions_updater
import octobot_trading.personal_data.positions as positions
import octobot_trading.enums as enums


class PositionsUpdaterSimulator(positions_updater.PositionsUpdater):
    def __init__(self, channel):
        super().__init__(channel)
        self.exchange_manager = None

    async def start(self):
        self.exchange_manager = self.channel.exchange_manager
        self.logger = logging.get_logger(f"{self.__class__.__name__}[{self.exchange_manager.exchange.name}]")
        await exchanges_channel.get_chan(constants.MARK_PRICE_CHANNEL, self.channel.exchange_manager.id) \
            .new_consumer(self.handle_mark_price)

    async def handle_mark_price(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str, mark_price):
        """
        MarkPrice channel consumer callback
        """
        try:
            await self._update_positions_status(cryptocurrency=cryptocurrency, symbol=symbol, mark_price=mark_price)
        except Exception as e:
            self.logger.exception(e, True, f"Fail to handle mark price : {e}")

    async def _update_positions_status(self, cryptocurrency: str, symbol: str, mark_price):
        """
        Ask positions to check their status
        Ask liquidation and P&L update process if required
        """
        for position in copy.copy(
                self.exchange_manager.exchange_personal_data.positions_manager.get_open_positions(symbol=symbol)):
            position_closed = False
            try:
                # ask orders to update their status
                async with position.lock:
                    position_closed = await self._update_position_status(position,
                                                                         mark_price)
            except Exception as e:
                raise e
            finally:
                # ensure always call fill callback
                if position_closed:
                    await exchanges_channel.get_chan(constants.POSITIONS_CHANNEL, self.channel.exchange_manager.id) \
                        .get_internal_producer().send(cryptocurrency=cryptocurrency,
                                                      symbol=position.symbol,
                                                      order=position.to_dict(),
                                                      is_from_bot=True,
                                                      is_liquidated=position.is_liquidated(),
                                                      is_closed=True,
                                                      is_updated=False)

    async def _update_position_status(self, position: positions.Position, mark_price):
        """
        Call position status update
        """
        position_closed = False
        try:
            await position.update_status(mark_price)

            if position.status in (enums.PositionStatus.CLOSED, enums.PositionStatus.LIQUIDATING):
                position_closed = True
                self.logger.debug(f"{position.symbol} (ID : {position.position_id})"
                                  f" closed on {self.channel.exchange.name} "
                                  f"at {position.mark_price} "
                                  f"by {'liquidation' if position.is_liquidated() else 'manual close'}")
                await position.close()
        except Exception as e:
            self.logger.exception(e, True, f"Fail to update position status : {e} (concerned position : {position}).")
        return position_closed



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import threading
import click
import click_shell

import octobot_trading.api as api
import octobot_trading.cli as cli
import octobot_trading.exchanges as exchanges
import octobot_trading.cli.cli_tools as cli_tools
import octobot_trading.enums as enums


@click_shell.shell(prompt='OctoBot-Trading > ', intro='Starting...')
def app():
    exchange_name = "binance"
    exchange_builder = exchanges.create_exchange_builder_instance(cli.get_config(), exchange_name) \
        .is_simulated() \
        .is_rest_only()

    cli.add_exchange(exchange_name, {
        "exchange_builder": exchange_builder,
        "exchange_thread": threading.Thread(target=cli_tools.start_cli_exchange, args=(exchange_builder,))
    })

    cli.get_exchange(exchange_name)["exchange_thread"].start()


@app.command()
def show():
    cli.set_should_display_callbacks_logs(True)


@app.command()
def hide():
    cli.set_should_display_callbacks_logs(False)


#  create_order --exchange_name binance --symbol BTC/USDT --price 11000 --quantity 1 --order_type buy_limit
@app.command()
@click.option("--exchange_name", prompt="Exchange name", help="The name of the exchange to use.", type=str)
@click.option("--symbol", prompt="Order symbol", help="The order symbol.", type=str)
@click.option("--price", prompt="Order price", help="The order price.", type=float)
@click.option("--quantity", prompt="Order quantity", help="The order quantity.", type=float)
@click.option("--order_type", prompt="Order type", help="The order type.",
              type=click.Choice([t.value for t in enums.TraderOrderType]))
def create_order(exchange_name, symbol, price, quantity, order_type):
    asyncio.get_event_loop().run_until_complete(
        api.create_order(cli.exchanges[exchange_name]["exchange_factory"].exchange_manager,
                         order_type=enums.TraderOrderType(order_type),
                         symbol=symbol,
                         current_price=price,
                         quantity=quantity,
                         price=price))


#  orders --exchange_name binance --symbol BTC/USDT
@app.command()
@click.option("--exchange_name", prompt="Exchange name", help="The name of the exchange to use.", type=str)
@click.option("--symbol", prompt="Order symbol", help="The order symbol.", type=str)
def orders(exchange_name, symbol):
    exchange_manager = cli.exchanges[exchange_name]["exchange_factory"].exchange_manager
    click.echo(api.get_open_orders(exchange_manager))


@app.command()
@click.option("--exchange_name", prompt="Exchange name", help="The name of the exchange to use.")
# @click.option("--api_key", prompt="API key", help="The api key of the exchange to use.")
# @click.option("--api_secret", prompt="API secret", help="The api secret of the exchange to use.")
def connect(exchange_name):
    if exchange_name not in cli.exchanges:
        exchange_builder = api.create_exchange_builder(cli.get_config(), exchange_name)

        cli.exchanges[exchange_name] = {
            "exchange_builder": exchange_builder,
            "exchange_thread": threading.Thread(target=cli_tools.start_cli_exchange, args=(exchange_builder,))
        }

        cli.exchanges[exchange_name]["exchange_thread"].start()
    else:
        click.echo("Already connected to this exchange", err=True)
        return
    click.echo(f"Connected to {exchange_name}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.orders import order_state
from octobot_trading.personal_data.orders.order_state import (
    OrderState,
)
from octobot_trading.personal_data.orders import order
from octobot_trading.personal_data.orders.order import (
    Order,
    parse_order_type,
)
from octobot_trading.personal_data.orders import types
from octobot_trading.personal_data.orders.types import (
    UnknownOrder,
    MarketOrder,
    SellMarketOrder,
    BuyMarketOrder,
    BuyLimitOrder,
    SellLimitOrder,
    LimitOrder,
    TakeProfitOrder,
    StopLossOrder,
    StopLossLimitOrder,
    TakeProfitLimitOrder,
    TrailingStopOrder,
    TrailingStopLimitOrder,
)
from octobot_trading.personal_data.orders import states
from octobot_trading.personal_data.orders.states import (
    CloseOrderState,
    CancelOrderState,
    OpenOrderState,
    create_order_state,
    FillOrderState,
)
from octobot_trading.personal_data.orders import channel
from octobot_trading.personal_data.orders.channel import (
    OrdersProducer,
    OrdersChannel,
    OrdersUpdater,
    OrdersUpdaterSimulator,
)
from octobot_trading.personal_data.orders import orders_manager
from octobot_trading.personal_data.orders.orders_manager import (
    OrdersManager,
)
from octobot_trading.personal_data.orders import order_util
from octobot_trading.personal_data.orders.order_util import (
    is_valid,
    get_min_max_amounts,
    check_cost,
    total_fees_from_order_dict,
    get_fees_for_currency,
    parse_order_status,
    parse_is_cancelled,
    get_pre_order_data,
)
from octobot_trading.personal_data.orders import order_adapter
from octobot_trading.personal_data.orders.order_adapter import (
    adapt_price,
    adapt_quantity,
    trunc_with_n_decimal_digits,
    adapt_order_quantity_because_quantity,
    adapt_order_quantity_because_price,
    split_orders,
    check_and_adapt_order_details_if_necessary,
    add_dusts_to_quantity_if_necessary,
)
from octobot_trading.personal_data.orders import order_factory
from octobot_trading.personal_data.orders.order_factory import (
    create_order_from_raw,
    create_order_instance_from_raw,
    create_order_from_type,
    create_order_instance,
)

__all__ = [
    "Order",
    "parse_order_type",
    "is_valid",
    "get_min_max_amounts",
    "check_cost",
    "total_fees_from_order_dict",
    "get_fees_for_currency",
    "parse_order_status",
    "parse_is_cancelled",
    "get_pre_order_data",
    "OrderState",
    "OrdersUpdater",
    "adapt_price",
    "adapt_quantity",
    "trunc_with_n_decimal_digits",
    "adapt_order_quantity_because_quantity",
    "adapt_order_quantity_because_price",
    "split_orders",
    "check_and_adapt_order_details_if_necessary",
    "add_dusts_to_quantity_if_necessary",
    "create_order_from_raw",
    "create_order_instance_from_raw",
    "create_order_from_type",
    "create_order_instance",
    "OrdersProducer",
    "OrdersChannel",
    "OrdersManager",
    "OrdersUpdaterSimulator",
    "CloseOrderState",
    "CancelOrderState",
    "OpenOrderState",
    "create_order_state",
    "FillOrderState",
    "UnknownOrder",
    "MarketOrder",
    "SellMarketOrder",
    "BuyMarketOrder",
    "BuyLimitOrder",
    "SellLimitOrder",
    "LimitOrder",
    "TakeProfitOrder",
    "StopLossOrder",
    "StopLossLimitOrder",
    "TakeProfitLimitOrder",
    "TrailingStopOrder",
    "TrailingStopLimitOrder",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

import octobot_commons.constants as commons_constants
from octobot_commons.tests.test_config import load_test_config
from octobot_trading.exchanges.traders.trader_simulator import TraderSimulator
from octobot_trading.exchanges.traders.trader import Trader


@pytest.fixture
async def trader(exchange_manager):
    config = load_test_config()
    config[commons_constants.CONFIG_TRADER][commons_constants.CONFIG_ENABLED_OPTION] = True
    trader_inst = Trader(load_test_config(), exchange_manager)
    await trader_inst.initialize()
    return config, exchange_manager, trader_inst


@pytest.fixture
async def trader_simulator(exchange_manager):
    config = load_test_config()
    config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_ENABLED_OPTION] = True
    trader_inst = TraderSimulator(load_test_config(), exchange_manager)
    await trader_inst.initialize()
    return config, exchange_manager, trader_inst



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.symbol_util as symbol_util
import octobot_commons.time_frame_manager as time_frame_manager
import octobot_commons.constants as constants
import octobot_commons.logging as logging

import octobot_trading.util as util


class ExchangeConfig(util.Initializable):
    def __init__(self, exchange_manager):
        super().__init__()
        self._logger = logging.get_logger(self.__class__.__name__)

        self.exchange_manager = exchange_manager
        self.config = exchange_manager.config

        # dict of exchange supported pairs by enabled currencies from self.config
        self.traded_cryptocurrencies = {}
        # list of exchange supported enabled pairs from self.config
        self.traded_symbol_pairs = []
        # list of exchange supported pairs from self.config
        self.all_config_symbol_pairs = []
        # list of exchange supported time frames
        self.traded_time_frames = []
        # list of time frames to be used for real-time purposes (short time frames)
        self.real_time_time_frames = []

    async def initialize_impl(self):
        pass

    def set_config_traded_pairs(self):  # TODO
        self._set_config_traded_pairs()

    def set_config_time_frame(self):  # TODO
        self._set_config_time_frame()

    def get_shortest_time_frame(self):
        return self.traded_time_frames[-1]

    async def handle_symbol_update(self, exchange: str, exchange_id: str, crypto_currency: str, symbols: list) -> tuple:
        try:
            return self._add_tradable_symbols(crypto_currency, symbols)
        except Exception as e:
            self._logger.exception(e, True, f"Fail to handle symbol update : {e}")

    async def handle_time_frame_update(self, exchange: str, exchange_id: str, time_frames: list) -> list:
        try:
            return self._add_tradable_time_frames(time_frames)
        except Exception as e:
            self._logger.exception(e, True, f"Fail to handle time frame update : {e}")

    def _set_config_traded_pairs(self):
        self.traded_cryptocurrencies = {}
        traded_symbol_pairs_set = set()
        existing_pairs = set()
        for cryptocurrency in self.config[constants.CONFIG_CRYPTO_CURRENCIES]:
            if self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_PAIRS]:
                is_enabled = util.is_currency_enabled(self.config, cryptocurrency, True)
                if self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_PAIRS] != \
                        constants.CONFIG_SYMBOLS_WILDCARD:
                    self._populate_non_wildcard_pairs(cryptocurrency, existing_pairs, is_enabled)
                else:
                    self._populate_wildcard_pairs(cryptocurrency, existing_pairs, is_enabled)
                # add to global traded pairs
                if is_enabled:
                    if not self.traded_cryptocurrencies[cryptocurrency]:
                        self._logger.error(
                            f"{self.exchange_manager.exchange_name} is not supporting any {cryptocurrency} trading pair"
                            f" from the current configuration.")
                    traded_symbol_pairs_set = traded_symbol_pairs_set.union(
                        self.traded_cryptocurrencies[cryptocurrency]
                    )
            else:
                self._logger.error(f"Current configuration for {cryptocurrency} is not including any trading pair, "
                                   f"this asset can't be traded and related orders won't be loaded. "
                                   f"OctoBot requires at least one trading pair in configuration to handle an asset. "
                                   f"You can add trading pair(s) for each asset in the configuration section.")
        self.traded_symbol_pairs = list(traded_symbol_pairs_set)
        self.all_config_symbol_pairs = list(existing_pairs)

    def _populate_non_wildcard_pairs(self, cryptocurrency, existing_pairs, is_enabled):
        if self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_PAIRS] != \
                constants.CONFIG_SYMBOLS_WILDCARD:
            currency_pairs = []
            for symbol in self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][
                constants.CONFIG_CRYPTO_PAIRS]:
                if self.exchange_manager.symbol_exists(symbol):
                    if is_enabled:
                        currency_pairs.append(symbol)
                    # also add disabled pairs to existing pairs since they still exist on exchange
                    existing_pairs.add(symbol)
                elif is_enabled:
                    self._logger.error(f"{self.exchange_manager.exchange_name} is not supporting the "
                                       f"{symbol} trading pair.")
            if is_enabled:
                self.traded_cryptocurrencies[cryptocurrency] = currency_pairs

    def _populate_wildcard_pairs(self, cryptocurrency, existing_pairs, is_enabled):
        wildcard_pairs_list = self._create_wildcard_symbol_list(self.config[constants.CONFIG_CRYPTO_CURRENCIES]
                                              [cryptocurrency][constants.CONFIG_CRYPTO_QUOTE])

        # additional pairs
        if constants.CONFIG_CRYPTO_ADD in self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency]:
            wildcard_pairs_list += self._add_tradable_symbols_from_config(cryptocurrency,
                                                                          wildcard_pairs_list)

        if is_enabled:
            self.traded_cryptocurrencies[cryptocurrency] = wildcard_pairs_list

        # also add disabled pairs to existing pairs since they still exist on exchange
        existing_pairs.update(wildcard_pairs_list)

    def _set_config_time_frame(self):
        for time_frame in time_frame_manager.get_config_time_frame(self.config):
            if self.exchange_manager.time_frame_exists(time_frame.value):
                self.traded_time_frames.append(time_frame)
        if not self.exchange_manager.is_backtesting or not self.traded_time_frames:
            # add shortest time frame for realtime evaluators or if no time frame at all has
            # been registered in backtesting
            client_shortest_time_frame = time_frame_manager.find_min_time_frame(
                self.exchange_manager.client_time_frames,
                constants.MIN_EVAL_TIME_FRAME)
            self.real_time_time_frames.append(client_shortest_time_frame)
        self.traded_time_frames = list(set().union(self.traded_time_frames, self.real_time_time_frames))
        self.traded_time_frames = time_frame_manager.sort_time_frames(self.traded_time_frames, reverse=True)

    @staticmethod
    def _is_tradable_with_cryptocurrency(symbol, cryptocurrency):
        return symbol if symbol_util.split_symbol(symbol)[1] == cryptocurrency else None

    def _add_tradable_symbols_from_config(self, cryptocurrency, filtered_symbols):
        return [
            symbol
            for symbol in self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_ADD]
            if self.exchange_manager.symbol_exists(symbol) and symbol not in filtered_symbols
        ]

    def _create_wildcard_symbol_list(self, cryptocurrency):
        return [s for s in [ExchangeConfig._is_tradable_with_cryptocurrency(symbol, cryptocurrency)
                            for symbol in self.exchange_manager.client_symbols]
                if s is not None]

    def _add_tradable_symbols(self, cryptocurrency, symbols):
        if cryptocurrency not in self.config[constants.CONFIG_CRYPTO_CURRENCIES]:
            # TODO use exchange config
            self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency] = {}
            self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_PAIRS] = symbols
            return cryptocurrency, symbols

        # TODO manage wildcard
        symbols_to_add = [s for s in symbols
                          if self.exchange_manager.symbol_exists(s)
                          and s not in self.config[constants.CONFIG_CRYPTO_CURRENCIES]
                          [cryptocurrency][constants.CONFIG_CRYPTO_PAIRS]]

        # TODO use exchange config
        self.config[constants.CONFIG_CRYPTO_CURRENCIES][cryptocurrency][constants.CONFIG_CRYPTO_PAIRS] += symbols_to_add

        return None, symbols_to_add

    def _add_tradable_time_frames(self, time_frames):
        # TODO use exchange config
        time_frames_to_add = [tf for tf in time_frames
                              if self.exchange_manager.time_frame_exists(tf)
                              and tf not in self.config[constants.CONFIG_TIME_FRAME]]
        return time_frames_to_add



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.logging as logging

import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.enums as enums
import octobot_trading.util as util
import octobot_trading.constants
import octobot_trading.errors


class OrderState(util.Initializable):
    def __init__(self, order, is_from_exchange_data):
        super().__init__()

        # ensure order has not been cleared
        if order.is_cleared():
            raise octobot_trading.errors.InvalidOrderState(f"Order has already been cleared. Order: {order}")

        # related order
        self.order = order

        # default state
        self.state = enums.OrderStates.UNKNOWN

        # if this state has been created from exchange data or OctoBot internal mechanism
        self.is_from_exchange_data = is_from_exchange_data

        # order state lock
        self.lock = asyncio.Lock()

    def is_pending(self) -> bool:
        """
        :return: True if the state is pending for update
        """
        return self.state is enums.OrderStates.UNKNOWN

    def is_refreshing(self) -> bool:
        """
        :return: True if the state is updating
        """
        return self.state is enums.OrderStates.REFRESHING

    def is_open(self) -> bool:
        """
        :return: True if the Order is considered as open
        """
        return not (self.is_filled() or self.is_canceled() or self.is_closed())

    def is_filled(self) -> bool:
        """
        :return: True if the Order is considered as filled
        """
        return False

    def is_closed(self) -> bool:
        """
        :return: True if the Order is considered as closed
        """
        return False

    def is_canceled(self) -> bool:
        """
        :return: True if the Order is considered as canceled
        """
        return False

    def get_logger(self):
        """
        :return: the order logger
        """
        return logging.get_logger(self.order.get_logger_name() if self.order is not None else
                                  f"{self.__class__.__name__}_without_order")

    def log_order_event_message(self, state_message):
        """
        Log an order state event
        """
        self.get_logger().info(f"{self.order} {state_message}"
                               f" on {self.order.exchange_manager.exchange_name}")

    async def initialize_impl(self) -> None:
        """
        Default OrderState initialization process
        """
        await self.update()

    async def terminate(self) -> None:
        """
        Implement the state ending process
        Can be portfolio updates, fees request, linked order cancellation, Trade creation etc...
        """
        raise NotImplementedError("terminate not implemented")

    async def update(self) -> None:
        """
        Update the order state if necessary
        Necessary when the state is not already synchronizing and when the order type is supported by the exchange
        Try to fix the pending state or terminate
        """
        if not self.is_refreshing():
            if self.is_pending() and not self.order.is_self_managed():
                self.log_order_event_message("synchronizing")
                await self.synchronize()
            else:
                async with self.lock:
                    await self.terminate()
        else:
            self.get_logger().debug(f"Trying to update a refreshing state for order: {self.order}")

    async def synchronize(self, force_synchronization=False, catch_exception=False) -> None:
        """
        Implement the exchange synchronization process
        Should begin by setting the state to REFRESHING
        Should end by :
        - calling terminate if the state is terminated
        - restoring the initial state if nothing has been changed with synchronization or if sync failed
        :param force_synchronization: When True, for the update of the order from the exchange
        :param catch_exception: When False raises the Exception during synchronize_order instead of catching it silently
        """
        try:
            await self._synchronize_order_with_exchange(force_synchronization=force_synchronization)
        except Exception as e:
            if catch_exception:
                self.get_logger().exception(e, True, f"Error when synchronizing order {self.order}: {e}")
            else:
                raise

    async def _refresh_order_from_exchange(self, force_synchronization=False) -> bool:
        """
        Ask OrdersChannel Internal producer to refresh the order from the exchange
        :param force_synchronization: When True, for the update of the order from the exchange
        :return: the result of OrdersProducer.update_order_from_exchange()
        """
        return (await exchange_channel.get_chan(octobot_trading.constants.ORDERS_CHANNEL,
                                                self.order.exchange_manager.id).get_internal_producer().
                update_order_from_exchange(order=self.order,
                                           wait_for_refresh=True,
                                           force_job_execution=force_synchronization))

    async def on_order_refresh_successful(self):
        """
        Called when _synchronize_order_with_exchange succeed to update the order
        """
        raise NotImplementedError("_on_order_refresh_successful not implemented")

    async def _synchronize_order_with_exchange(self, force_synchronization: bool = False) -> None:
        """
        Ask the exchange to update the order only if the state is not already refreshing
        When the refreshing process starts set the state to enums.OrderStates.REFRESHING
        :param force_synchronization: When True, for the update of the order from the exchange
        """
        if self.is_refreshing():
            self.get_logger().debug(f"Trying to refresh a refreshing state for order: {self.order}")
        else:
            previous_state = self.state
            async with self.lock:
                self.state = enums.OrderStates.REFRESHING
            await self._refresh_or_restore_order_from_exchange(
                previous_state=previous_state,
                force_synchronization=force_synchronization)

    async def _refresh_or_restore_order_from_exchange(self,
                                                      previous_state: enums.OrderStates,
                                                      force_synchronization: bool = False) -> None:
        """
        Ask the exchange to update the order
        Restore the previous state if the refresh process fails
        :param force_synchronization: When True, for the update of the order from the exchange
        """
        try:
            await self._refresh_order_from_exchange(force_synchronization=force_synchronization)
        finally:
            async with self.lock:
                if self.state is enums.OrderStates.REFRESHING:
                    self.state = previous_state

    def clear(self):
        """
        Clear references
        """
        self.order = None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library

import octobot_trading.enums as enums

# Strings
CURRENT_PORTFOLIO_STRING = "Current Portfolio :"
CONFIG_PORTFOLIO_INFO = "info"
CONFIG_PORTFOLIO_FREE = "free"
CONFIG_PORTFOLIO_USED = "used"
CONFIG_PORTFOLIO_TOTAL = "total"
CONFIG_PORTFOLIO_MARGIN = "margin"
REAL_TRADER_STR = "[Real Trader] "
SIMULATOR_TRADER_STR = "[Simulator] "

# Trader
DEFAULT_REFERENCE_MARKET = "BTC"
CURRENCY_DEFAULT_MAX_PRICE_DIGITS = 8

# Order creation
ORDER_DATA_FETCHING_TIMEOUT = 60

# Tentacles
TRADING_MODE_REQUIRED_STRATEGIES = "required_strategies"
TRADING_MODE_REQUIRED_STRATEGIES_MIN_COUNT = "required_strategies_min_count"
TENTACLES_TRADING_MODE_PATH = "Mode"

# Exchange
DEFAULT_EXCHANGE_TIME_LAG = 10
DEFAULT_BACKTESTING_TIME_LAG = 0
INFINITE_MAX_HANDLED_PAIRS_WITH_TIMEFRAME = -1

TESTED_EXCHANGES = ["binance", "bitmax", "kucoin", "coinbasepro"]
SIMULATOR_TESTED_EXCHANGES = ["bybit"]

CONFIG_DEFAULT_FEES = 0.1
CONFIG_DEFAULT_SIMULATOR_FEES = 0

# API
API_LOGGER_TAG = "TradingApi"

# Channels
# Exchange public data
TICKER_CHANNEL = "Ticker"
MINI_TICKER_CHANNEL = "MiniTicker"
RECENT_TRADES_CHANNEL = "RecentTrade"
LIQUIDATIONS_CHANNEL = "Liquidations"
ORDER_BOOK_CHANNEL = "OrderBook"
ORDER_BOOK_TICKER_CHANNEL = "OrderBookTicker"
KLINE_CHANNEL = "Kline"
OHLCV_CHANNEL = "OHLCV"
MARK_PRICE_CHANNEL = "MarkPrice"
FUNDING_CHANNEL = "Funding"

# Exchange personal data
TRADES_CHANNEL = "Trades"
ORDERS_CHANNEL = "Orders"
BALANCE_CHANNEL = "Balance"
BALANCE_PROFITABILITY_CHANNEL = "BalanceProfitability"
POSITIONS_CHANNEL = "Positions"

# 946742400 is 01/01/2000, if trade time is lower, there is an issue.
MINIMUM_VAL_TRADE_TIME = 946688400

# Internal
MODE_CHANNEL = "Mode"

# CCXT library constants
CCXT_INFO = "info"

WEBSOCKET_FEEDS_TO_TRADING_CHANNELS = {
    TICKER_CHANNEL: [enums.WebsocketFeeds.TICKER],
    MINI_TICKER_CHANNEL: [enums.WebsocketFeeds.MINI_TICKER],
    RECENT_TRADES_CHANNEL: [enums.WebsocketFeeds.TRADES],
    LIQUIDATIONS_CHANNEL: [enums.WebsocketFeeds.LIQUIDATIONS],
    ORDER_BOOK_CHANNEL: [enums.WebsocketFeeds.L2_BOOK, enums.WebsocketFeeds.L3_BOOK],
    ORDER_BOOK_TICKER_CHANNEL: [enums.WebsocketFeeds.BOOK_TICKER],
    KLINE_CHANNEL: [enums.WebsocketFeeds.KLINE],
    OHLCV_CHANNEL: [enums.WebsocketFeeds.CANDLE],
    TRADES_CHANNEL: [enums.WebsocketFeeds.TRADE],
    ORDERS_CHANNEL: [enums.WebsocketFeeds.ORDERS],
    MARK_PRICE_CHANNEL: [enums.WebsocketFeeds.MARK_PRICE],
    BALANCE_CHANNEL: [enums.WebsocketFeeds.PORTFOLIO],
    POSITIONS_CHANNEL: [enums.WebsocketFeeds.POSITION],
    FUNDING_CHANNEL: [enums.WebsocketFeeds.FUNDING]
}

FILL_ORDER_STATUS_SCOPE = [enums.OrderStatus.CLOSED,
                           enums.OrderStatus.FILLED,
                           enums.OrderStatus.PARTIALLY_FILLED]
CANCEL_ORDER_STATUS_SCOPE = [enums.OrderStatus.PENDING_CANCEL,
                             enums.OrderStatus.CANCELED,
                             enums.OrderStatus.EXPIRED,
                             enums.OrderStatus.REJECTED]
